/* AP 05-04-04	gocc_chan.c				*/
/*		channel construct for GridOccam		*/

# include <pthread.h>
# include <stdio.h>

# include "gocc_chan.h"

# ifndef NULL
# define NULL (void *) 0
# endif

typedef struct _chan_struct {
	pthread_mutex_t * mutex;
	pthread_cond_t * send_cond;
	pthread_cond_t * recv_cond;
	pthread_cond_t * recv_finished;
	int link_cnt;
	int data_valid;
	int data;
} chan_struct, * chan_p;

static chan_p channels[MAX_CHAN];

static chan_p alloc_chan( int id ) {
	chan_p cp;

	if ((id >= MAX_CHAN) || (id < 0)) return NULL;
	if (channels[id] != NULL) return channels[id];

	channels[id] = cp = (chan_p) malloc( sizeof(chan_struct));
	cp->mutex = (pthread_mutex_t *) malloc( sizeof( pthread_mutex_t ));
	pthread_mutex_init( cp->mutex, NULL );

	cp->send_cond = (pthread_cond_t *) malloc( sizeof( pthread_cond_t ));
	cp->recv_cond = (pthread_cond_t *) malloc( sizeof( pthread_cond_t ));
	cp->recv_finished = (pthread_cond_t *) malloc( sizeof( pthread_cond_t ));
	pthread_cond_init( cp->send_cond, NULL );
	pthread_cond_init( cp->recv_cond, NULL );
	pthread_cond_init( cp->recv_finished, NULL );
	cp->link_cnt = 0;
	cp->data_valid = 0;
	return cp;
}

static int free_chan( int id ) {
	chan_p cp = channels[id];
	if (!cp) return -1;
	pthread_mutex_destroy( cp->mutex );
	pthread_cond_destroy( cp->send_cond );
	pthread_cond_destroy( cp->recv_cond );
	free( cp->mutex );
	free( cp->send_cond );
	free( cp->recv_cond );
	free( cp );
	channels[id] = NULL;
	return 0;
}

int conn_chan( int id ) {
	chan_p cp = alloc_chan( id );
	if (!cp) return -1;
	cp->link_cnt++;
	return 0;
}

int break_chan( int id ) {
	chan_p cp;
	if ((id >= MAX_CHAN) || (id < 0)) return -1;
	if(!(cp = channels[id])) return -2;
	cp->link_cnt--;
	if (cp->link_cnt <= 0) free_chan( id );
	return 0;
}

int send_chan( int id, int data ) {
	chan_p cp;
	if ((id >= MAX_CHAN) || (id < 0)) return -1;
	if(!(cp = channels[id])) return -2;

	pthread_mutex_lock( cp->mutex );
	while (cp->data_valid) 
		pthread_cond_wait( cp->send_cond, cp->mutex );
	cp->data_valid = 1;
	cp->data = data;
# ifdef _DEBUG
	printf("send_chan( id = %d, val = %d )\t -- caller is %d\n",
                id, data, pthread_self() ); fflush(stdout);
# endif
	pthread_cond_signal( cp->recv_cond );
	pthread_cond_wait( cp->recv_finished, cp->mutex );
	pthread_mutex_unlock( cp->mutex );
}

int recv_chan( int id, int * data ) {
	chan_p cp;
	if ((id >= MAX_CHAN) || (id < 0)) return -1;
	if(!(cp = channels[id])) return -2;

	pthread_mutex_lock( cp->mutex );
	while (!cp->data_valid)
		pthread_cond_wait( cp->recv_cond, cp->mutex );
	cp->data_valid = 0;
	*data = cp->data;
# ifdef _DEBUG
	printf("recv_chan( id = %d, val = %d )\t -- caller is %d\n",
                id, * data, pthread_self() ); fflush(stdout);
# endif
	pthread_cond_signal( cp->recv_finished );
	pthread_mutex_unlock( cp->mutex );
	pthread_cond_signal( cp->send_cond );
}

